@@ -5,7 +5,7 @@ package OpenAPI::Modern;
 # ABSTRACT: Validate HTTP requests and responses against an OpenAPI document
 # KEYWORDS: validation evaluation JSON Schema OpenAPI Swagger HTTP request response
 
-our $VERSION = '0.005';
+our $VERSION = '0.006';
 
 use 5.020;  # for fc, unicode_strings features
 use Moo;
@@ -18,8 +18,9 @@ no if "$]" >= 5.033006, feature => 'bareword_filehandles';
 use Carp 'croak';
 use Safe::Isa;
 use Ref::Util qw(is_plain_hashref is_plain_arrayref);
-use List::Util 'first';
+use List::Util 1.45 qw(first uniq);
 use Scalar::Util 'looks_like_number';
+use Module::Runtime 'require_module';
 use Feature::Compat::Try;
 use JSON::Schema::Modern 0.527;
 use JSON::Schema::Modern::Utilities qw(jsonp canonical_uri E abort get_type);
@@ -48,6 +49,10 @@ has evaluator => (
   handles => [ qw(get_media_type add_media_type) ],
 );
 
+sub load_vocabulary_classes ($self) {
+  require_module($_) foreach uniq map $_->{vocabularies}->@*, $self->evaluator->_resource_values;
+}
+
 around BUILDARGS => sub ($orig, $class, @args) {
   my $args = $class->$orig(@args);
 
@@ -247,21 +252,14 @@ sub validate_response ($self, $response, $options) {
 
 # for now, we only use captures, rather than parsing the URI directly.
 sub _validate_path_parameter ($self, $state, $param_obj, $captures) {
-  return E({ %$state, keyword => 'content' }, 'content not yet supported')
-    if exists $param_obj->{content};
-
   # 'required' is always true for path parameters
   return E({ %$state, keyword => 'required' }, 'missing path parameter: %s', $param_obj->{name})
     if not exists $captures->{$param_obj->{name}};
 
-  $state = { %$state, schema_path => jsonp($state->{schema_path}, 'schema') };
-  $self->_evaluate_subschema($captures->{$param_obj->{name}}, $param_obj->{schema}, $state);
+  $self->_validate_parameter_content($state, $param_obj, \ $captures->{$param_obj->{name}});
 }
 
 sub _validate_query_parameter ($self, $state, $param_obj, $uri) {
-  return E({ %$state, keyword => 'content' }, 'content not yet supported')
-    if exists $param_obj->{content};
-
   # parse the query parameters out of uri
   my $query_params = { $uri->query_form };
 
@@ -278,15 +276,11 @@ sub _validate_query_parameter ($self, $state, $param_obj, $uri) {
 
   # TODO: check 'allowReserved': it cannot be supported if we use proper URL encoding
 
-  $state = { %$state, schema_path => jsonp($state->{schema_path}, 'schema') };
-  $self->_evaluate_subschema($query_params->{$param_obj->{name}}, $param_obj->{schema}, $state);
+  $self->_validate_parameter_content($state, $param_obj, \ $query_params->{$param_obj->{name}});
 }
 
 # validates a header, from either the request or the response
 sub _validate_header_parameter ($self, $state, $header_name, $header_obj, $headers) {
-  return E({ %$state, keyword => 'content' }, 'content not yet supported')
-    if exists $header_obj->{content};
-
   # NOTE: for now, we will only support a single value.
   my @values = $headers->header($header_name);
   if (not @values) {
@@ -295,14 +289,38 @@ sub _validate_header_parameter ($self, $state, $header_name, $header_obj, $heade
     return 1;
   }
 
-  $state = { %$state, schema_path => jsonp($state->{schema_path}, 'schema') };
-  $self->_evaluate_subschema($values[0], $header_obj->{schema}, $state);
+  $self->_validate_parameter_content($state, $header_obj, \ $values[0]);
 }
 
 sub _validate_cookie_parameter ($self, $state, $param_obj, $request) {
   return E($state, 'cookie parameters not yet supported');
 }
 
+sub _validate_parameter_content ($self, $state, $param_obj, $content_ref) {
+  if (exists $param_obj->{content}) {
+    my ($content_type) = keys $param_obj->{content}->%*;
+
+    my $media_type_decoder = $self->get_media_type($content_type);
+    abort({ %$state, keyword => 'content', _schema_path_suffix => $content_type },
+        'EXCEPTION: unsupported Content-Type "%s": add support with $openapi->add_media_type(...)', $content_type)
+      if not $media_type_decoder;
+
+    try {
+      $content_ref = $media_type_decoder->($content_ref);
+    }
+    catch ($e) {
+      return E({ %$state, keyword => 'content', _schema_path_suffix => $content_type },
+        'could not decode content as %s', $content_type);
+    }
+
+    $state = { %$state, schema_path => jsonp($state->{schema_path}, 'content', $content_type, 'schema') };
+    return $self->_evaluate_subschema($content_ref->$*, $param_obj->{content}{$content_type}{schema}, $state);
+  }
+
+  $state = { %$state, schema_path => jsonp($state->{schema_path}, 'schema') };
+  $self->_evaluate_subschema($content_ref->$*, $param_obj->{schema}, $state);
+}
+
 sub _validate_body_content ($self, $state, $content_obj, $message) {
   my $content_type = $message->content_type;
 
@@ -330,14 +348,27 @@ sub _validate_body_content ($self, $state, $content_obj, $message) {
 
   # decode the charset
   my $charset = $message->content_charset;
-  $decoded_content_ref = \ Encode::decode($charset, $decoded_content_ref->$*, Encode::FB_CROAK | Encode::LEAVE_SRC);
+  try {
+    $decoded_content_ref =
+      \ Encode::decode($charset, $decoded_content_ref->$*, Encode::FB_CROAK | Encode::LEAVE_SRC);
+  }
+  catch ($e) {
+    return E({ %$state, keyword => 'content', _schema_path_suffix => $content_type },
+      'could not decode content as %s', $charset);
+  }
 
   my $media_type_decoder = $self->get_media_type($content_type);
   abort({ %$state, keyword => 'content', _schema_path_suffix => $content_type },
       'EXCEPTION: unsupported Content-Type "%s": add support with $openapi->add_media_type(...)', $content_type)
     if not $media_type_decoder;
 
-  $decoded_content_ref = $media_type_decoder->($decoded_content_ref);
+  try {
+    $decoded_content_ref = $media_type_decoder->($decoded_content_ref);
+  }
+  catch ($e) {
+    return E({ %$state, keyword => 'content', _schema_path_suffix => $content_type },
+      'could not decode content as %s', $content_type);
+  }
 
   $state = { %$state, schema_path => jsonp($state->{schema_path}, 'content', $content_type, 'schema') };
   $self->_evaluate_subschema($decoded_content_ref->$*, $content_obj->{$content_type}{schema}, $state);
@@ -350,7 +381,7 @@ sub _result ($self, $state) {
     valid => !$state->{errors}->@*,
     !$state->{errors}->@*
       ? ($self->evaluator->collect_annotations
-        ? (annotations => $state->{annotations}) : ())
+        ? (annotations => $state->{annotations}//[]) : ())
       : (errors => $state->{errors}),
   );
 }
@@ -415,7 +446,7 @@ OpenAPI::Modern - Validate HTTP requests and responses against an OpenAPI docume
 
 =head1 VERSION
 
-version 0.005
+version 0.006
 
 =head1 SYNOPSIS
 
@@ -601,6 +632,12 @@ C<path_template>: a string representing the request URI, with placeholders in br
 
 More options will be added later, providing more flexible matching of the document to the request.
 
+=head2 load_vocabulary_classes
+
+Loads all vocabulary classes found within the resource index of the JSON Schema evaluator.
+Useful when deserializing an object from disk (where resource data exists in memory but the
+corresponding implementations are no longer loaded.
+
 =head1 ON THE USE OF JSON SCHEMAS
 
 Embedded JSON Schemas, through the use of the C<schema> keyword, are fully draft2020-12-compliant,
@@ -656,10 +693,6 @@ cookie parameters are not checked at all yet
 
 for query and header parameters, only the first value of each name is considered
 
-=item *
-
-media-type encodings in parameters are not yet supported
-
 =back
 
 =head1 SEE ALSO
